/**************************************************************************
* DSemu - The Next Generation                                             *
* Portable ARM7TDMI core: Class definition [arm7tdmi.h]                   *
* Copyright Imran Nazar, 2005; released under the BSD public licence.     *
**************************************************************************/

#ifndef __ARM7TDMI_H_
#define __ARM7TDMI_H_

#include <utility>
#include <map>
#include <stdio.h>
#include "defs.h"
#include "plgcpu.h"
#include "plgmmu32.h"
#include "plggui.h"
#include "datadefs.h"
#include "armdasm.h"
#include "codeblock.h"

// An implementation of the CPU plugin interface.
class ARM7TDMI : public CPUPlugin
{
    public:
        ARM7TDMI(std::string name,REQPTR req, UNREQPTR unreq);
        ~ARM7TDMI();

        PLUGININFO *getinfo();
        void reset();
        void release();
        void status(int,int);

        int exec(int);
	void clockAdd(int);
	void interrupt(int);
	
	void dbgBkptToggle(u32);
	std::map<u32, int> dbgBkptGet();

	void setPC(u32);

    private:

        //---Plugin-generic stuff
        std::string pName, pClass;
        REQPTR pRequest;
        UNREQPTR pUnrequest;
	static PLUGININFO pInfo;
        
        static std::string pluginName;

	//---Start of the stuff we'll be needing
        static MMU32Plugin *MMU;
	static GUIPlugin *GUI;
	
	static u32 *dbgbuffer;
	static int dbgwinID;
        static ARMDasm dasm;
	static u32 rdWhelper(u32);

        static FILE *dumpfile;
	
        //---Register set
        typedef struct {
            u32 r[16], flags[8], cpsr, spsr[7];
            u32 tmp1, tmp2, tmp3, tmp4;
	    uint64_t tmp64_1, tmp64_2;
	    u32 tmpc;
            u32 curop, curmode;
	    
	    u32 blockstart;
	    CODEBLOCK *curblock;

            u32 r0,r1,r2,r3,r4,r5,r6,r7,r8,r9,r10,r11,r12,r13,r14,r15;
            u32 r8fiq, r9fiq, r10fiq, r11fiq, r12fiq, r13fiq, r14fiq;
            u32 r13svc, r14svc, r13abt, r14abt, r13irq, r14irq, r13und, r14und;

            uint64_t clkcount;
        } REGS;

        static REGS reg;

        static std::map<u32, int> bkpts;

	//---Internal flag to/from CPSR helpers
        static void cpsrSplit(), cpsrUpdate();
	
        //---Opcode tables
        typedef int OPCODE;
        typedef OPCODE (*opfptr)();
	
        static const opfptr OLUT[4096];
	static const opfptr TLUT[256];

#ifdef ARM7TDMI_X86	
        static const opfptr XOLUT[4096];
	static const opfptr XTLUT[256];
	static void translate(CODEBLOCK*);
#endif
	
	//---Mode-switching
	static const char *modeStrings[7];
	static const int modeToCpsrLUT[7];
	static const int modeFromCpsrLUT[16];
	static void modeSwitch(int, int);

        //---Block handling
	static std::map<u32, CODEBLOCK> blocks;
	static int execOne(uint64_t);
	
        //---Conditional execution helpers
	static int condEQ(), condNE(), condCS(), condCC();
	static int condMI(), condPL(), condVS(), condVC();
	static int condHI(), condLS(), condGE(), condLT();
	static int condGT(), condLE(), condAL(), condNV();
	
	typedef int (*intfptr)();
	static const intfptr condLUT[16];
	
	//---Helper functions
	static u32 nlpo2(u32);
	static int ssum16(u16);

	//---Exceptional opcodes
	static OPCODE opUNP();   // Unpredictable op
        static OPCODE opUND();   // Undefined op
        static OPCODE opUNI();   // Unimplemented op
	static OPCODE opUNL();   // Unlinked (opcode cache filler)
        static OPCODE topUND();  // Thumb-mode undefined
	
        //---Instructions

        //---Branch (with Link)----------------------------------------------------

        static OPCODE opB(), opBL();
        static OPCODE opBX();

        //---DP (DataProcessing); LsL/LsR/AsR/RoR; Immediate/Register--------------

        static OPCODE opANDlli(), opANDllr();
        static OPCODE opANDlri(), opANDlrr();
        static OPCODE opANDari(), opANDarr();
        static OPCODE opANDrri(), opANDrrr();
        static OPCODE opANDimm();

        static OPCODE opEORlli(), opEORllr();
        static OPCODE opEORlri(), opEORlrr();
        static OPCODE opEORari(), opEORarr();
        static OPCODE opEORrri(), opEORrrr();
        static OPCODE opEORimm();

        static OPCODE opSUBlli(), opSUBllr();
        static OPCODE opSUBlri(), opSUBlrr();
        static OPCODE opSUBari(), opSUBarr();
        static OPCODE opSUBrri(), opSUBrrr();
        static OPCODE opSUBimm();

        static OPCODE opRSBlli(), opRSBllr();
        static OPCODE opRSBlri(), opRSBlrr();
        static OPCODE opRSBari(), opRSBarr();
        static OPCODE opRSBrri(), opRSBrrr();
        static OPCODE opRSBimm();

        static OPCODE opADDlli(), opADDllr();
        static OPCODE opADDlri(), opADDlrr();
        static OPCODE opADDari(), opADDarr();
        static OPCODE opADDrri(), opADDrrr();
        static OPCODE opADDimm();

        static OPCODE opADClli(), opADCllr();
        static OPCODE opADClri(), opADClrr();
        static OPCODE opADCari(), opADCarr();
        static OPCODE opADCrri(), opADCrrr();
        static OPCODE opADCimm();

        static OPCODE opSBClli(), opSBCllr();
        static OPCODE opSBClri(), opSBClrr();
        static OPCODE opSBCari(), opSBCarr();
        static OPCODE opSBCrri(), opSBCrrr();
        static OPCODE opSBCimm();

        static OPCODE opRSClli(), opRSCllr();
        static OPCODE opRSClri(), opRSClrr();
        static OPCODE opRSCari(), opRSCarr();
        static OPCODE opRSCrri(), opRSCrrr();
        static OPCODE opRSCimm();

        static OPCODE opORRlli(), opORRllr();
        static OPCODE opORRlri(), opORRlrr();
        static OPCODE opORRari(), opORRarr();
        static OPCODE opORRrri(), opORRrrr();
        static OPCODE opORRimm();

        static OPCODE opMOVlli(), opMOVllr();
        static OPCODE opMOVlri(), opMOVlrr();
        static OPCODE opMOVari(), opMOVarr();
        static OPCODE opMOVrri(), opMOVrrr();
        static OPCODE opMOVimm();

        static OPCODE opBIClli(), opBICllr();
        static OPCODE opBIClri(), opBIClrr();
        static OPCODE opBICari(), opBICarr();
        static OPCODE opBICrri(), opBICrrr();
        static OPCODE opBICimm();

        static OPCODE opMVNlli(), opMVNllr();
        static OPCODE opMVNlri(), opMVNlrr();
        static OPCODE opMVNari(), opMVNarr();
        static OPCODE opMVNrri(), opMVNrrr();
        static OPCODE opMVNimm();

        static OPCODE opANDSlli(), opANDSllr();
        static OPCODE opANDSlri(), opANDSlrr();
        static OPCODE opANDSari(), opANDSarr();
        static OPCODE opANDSrri(), opANDSrrr();
        static OPCODE opANDSimm();

        static OPCODE opEORSlli(), opEORSllr();
        static OPCODE opEORSlri(), opEORSlrr();
        static OPCODE opEORSari(), opEORSarr();
        static OPCODE opEORSrri(), opEORSrrr();
        static OPCODE opEORSimm();

        static OPCODE opSUBSlli(), opSUBSllr();
        static OPCODE opSUBSlri(), opSUBSlrr();
        static OPCODE opSUBSari(), opSUBSarr();
        static OPCODE opSUBSrri(), opSUBSrrr();
        static OPCODE opSUBSimm();

        static OPCODE opRSBSlli(), opRSBSllr();
        static OPCODE opRSBSlri(), opRSBSlrr();
        static OPCODE opRSBSari(), opRSBSarr();
        static OPCODE opRSBSrri(), opRSBSrrr();
        static OPCODE opRSBSimm();

        static OPCODE opADDSlli(), opADDSllr();
        static OPCODE opADDSlri(), opADDSlrr();
        static OPCODE opADDSari(), opADDSarr();
        static OPCODE opADDSrri(), opADDSrrr();
        static OPCODE opADDSimm();

        static OPCODE opADCSlli(), opADCSllr();
        static OPCODE opADCSlri(), opADCSlrr();
        static OPCODE opADCSari(), opADCSarr();
        static OPCODE opADCSrri(), opADCSrrr();
        static OPCODE opADCSimm();

        static OPCODE opSBCSlli(), opSBCSllr();
        static OPCODE opSBCSlri(), opSBCSlrr();
        static OPCODE opSBCSari(), opSBCSarr();
        static OPCODE opSBCSrri(), opSBCSrrr();
        static OPCODE opSBCSimm();

        static OPCODE opRSCSlli(), opRSCSllr();
        static OPCODE opRSCSlri(), opRSCSlrr();
        static OPCODE opRSCSari(), opRSCSarr();
        static OPCODE opRSCSrri(), opRSCSrrr();
        static OPCODE opRSCSimm();

        static OPCODE opTSTSlli(), opTSTSllr();
        static OPCODE opTSTSlri(), opTSTSlrr();
        static OPCODE opTSTSari(), opTSTSarr();
        static OPCODE opTSTSrri(), opTSTSrrr();
        static OPCODE opTSTSimm();

        static OPCODE opTEQSlli(), opTEQSllr();
        static OPCODE opTEQSlri(), opTEQSlrr();
        static OPCODE opTEQSari(), opTEQSarr();
        static OPCODE opTEQSrri(), opTEQSrrr();
        static OPCODE opTEQSimm();

        static OPCODE opCMPSlli(), opCMPSllr();
        static OPCODE opCMPSlri(), opCMPSlrr();
        static OPCODE opCMPSari(), opCMPSarr();
        static OPCODE opCMPSrri(), opCMPSrrr();
        static OPCODE opCMPSimm();

        static OPCODE opCMNSlli(), opCMNSllr();
        static OPCODE opCMNSlri(), opCMNSlrr();
        static OPCODE opCMNSari(), opCMNSarr();
        static OPCODE opCMNSrri(), opCMNSrrr();
        static OPCODE opCMNSimm();

        static OPCODE opORRSlli(), opORRSllr();
        static OPCODE opORRSlri(), opORRSlrr();
        static OPCODE opORRSari(), opORRSarr();
        static OPCODE opORRSrri(), opORRSrrr();
        static OPCODE opORRSimm();

        static OPCODE opMOVSlli(), opMOVSllr();
        static OPCODE opMOVSlri(), opMOVSlrr();
        static OPCODE opMOVSari(), opMOVSarr();
        static OPCODE opMOVSrri(), opMOVSrrr();
        static OPCODE opMOVSimm();

        static OPCODE opBICSlli(), opBICSllr();
        static OPCODE opBICSlri(), opBICSlrr();
        static OPCODE opBICSari(), opBICSarr();
        static OPCODE opBICSrri(), opBICSrrr();
        static OPCODE opBICSimm();

        static OPCODE opMVNSlli(), opMVNSllr();
        static OPCODE opMVNSlri(), opMVNSlrr();
        static OPCODE opMVNSari(), opMVNSarr();
        static OPCODE opMVNSrri(), opMVNSrrr();
        static OPCODE opMVNSimm();

        //---Multiply/LongMul------------------------------------------------------

        static OPCODE opMUL(), opMLA();
        static OPCODE opUMULL(), opUMLAL(), opSMULL(), opSMLAL();

        static OPCODE opMULS(), opMLAS();
        static OPCODE opUMULLS(), opUMLALS(), opSMULLS(), opSMLALS();

        //---Move Register From Status; Imm/Reg; Cpsr/Spsr-------------------------

        static OPCODE opMRSrc(), opMRSrs();

        //---Move Status from Register; Cpsr/Spsr----------------------------------

        static OPCODE opMSRrc(), opMSRrs();
        static OPCODE opMSRic(), opMSRis();

        //---LS (Load/Store); PReidx/OFfset/PsTidx; Imm/Reg; Minus/Plus; (Shift)---

        static OPCODE opLDRptim(), opLDRofim(), opLDRprim();
        static OPCODE opLDRptip(), opLDRofip(), opLDRprip();

        static OPCODE opSTRptim(), opSTRofim(), opSTRprim();
        static OPCODE opSTRptip(), opSTRofip(), opSTRprip();

        static OPCODE opLDRBptim(), opLDRBofim(), opLDRBprim();
        static OPCODE opLDRBptip(), opLDRBofip(), opLDRBprip();

        static OPCODE opSTRBptim(), opSTRBofim(), opSTRBprim();
        static OPCODE opSTRBptip(), opSTRBofip(), opSTRBprip();

        static OPCODE opLDRHptim(), opLDRHofim(), opLDRHprim();
        static OPCODE opLDRHptip(), opLDRHofip(), opLDRHprip();
        static OPCODE opLDRHptrm(), opLDRHofrm(), opLDRHprrm();
        static OPCODE opLDRHptrp(), opLDRHofrp(), opLDRHprrp();

        static OPCODE opSTRHptim(), opSTRHofim(), opSTRHprim();
        static OPCODE opSTRHptip(), opSTRHofip(), opSTRHprip();
        static OPCODE opSTRHptrm(), opSTRHofrm(), opSTRHprrm();
        static OPCODE opSTRHptrp(), opSTRHofrp(), opSTRHprrp();

        static OPCODE opLDRTptim(), opLDRTofim(), opLDRTprim();
        static OPCODE opLDRTptip(), opLDRTofip(), opLDRTprip();

        static OPCODE opSTRTptim(), opSTRTofim(), opSTRTprim();
        static OPCODE opSTRTptip(), opSTRTofip(), opSTRTprip();

        static OPCODE opLDRBTptim(), opLDRBTofim(), opLDRBTprim();
        static OPCODE opLDRBTptip(), opLDRBTofip(), opLDRBTprip();

        static OPCODE opSTRBTptim(), opSTRBTofim(), opSTRBTprim();
        static OPCODE opSTRBTptip(), opSTRBTofip(), opSTRBTprip();

        static OPCODE opLDRSHptim(), opLDRSHofim(), opLDRSHprim();
        static OPCODE opLDRSHptip(), opLDRSHofip(), opLDRSHprip();
        static OPCODE opLDRSHptrm(), opLDRSHofrm(), opLDRSHprrm();
        static OPCODE opLDRSHptrp(), opLDRSHofrp(), opLDRSHprrp();

        static OPCODE opLDRSBptim(), opLDRSBofim(), opLDRSBprim();
        static OPCODE opLDRSBptip(), opLDRSBofip(), opLDRSBprip();
        static OPCODE opLDRSBptrm(), opLDRSBofrm(), opLDRSBprrm();
        static OPCODE opLDRSBptrp(), opLDRSBofrp(), opLDRSBprrp();

        static OPCODE opLDRptrmll(), opLDRofrmll(), opLDRprrmll();
        static OPCODE opLDRptrpll(), opLDRofrpll(), opLDRprrpll();
        static OPCODE opLDRptrmlr(), opLDRofrmlr(), opLDRprrmlr();
        static OPCODE opLDRptrplr(), opLDRofrplr(), opLDRprrplr();
        static OPCODE opLDRptrmar(), opLDRofrmar(), opLDRprrmar();
        static OPCODE opLDRptrpar(), opLDRofrpar(), opLDRprrpar();
        static OPCODE opLDRptrmrr(), opLDRofrmrr(), opLDRprrmrr();
        static OPCODE opLDRptrprr(), opLDRofrprr(), opLDRprrprr();

        static OPCODE opSTRptrmll(), opSTRofrmll(), opSTRprrmll();
        static OPCODE opSTRptrpll(), opSTRofrpll(), opSTRprrpll();
        static OPCODE opSTRptrmlr(), opSTRofrmlr(), opSTRprrmlr();
        static OPCODE opSTRptrplr(), opSTRofrplr(), opSTRprrplr();
        static OPCODE opSTRptrmar(), opSTRofrmar(), opSTRprrmar();
        static OPCODE opSTRptrpar(), opSTRofrpar(), opSTRprrpar();
        static OPCODE opSTRptrmrr(), opSTRofrmrr(), opSTRprrmrr();
        static OPCODE opSTRptrprr(), opSTRofrprr(), opSTRprrprr();

        static OPCODE opLDRBptrmll(), opLDRBofrmll(), opLDRBprrmll();
        static OPCODE opLDRBptrpll(), opLDRBofrpll(), opLDRBprrpll();
        static OPCODE opLDRBptrmlr(), opLDRBofrmlr(), opLDRBprrmlr();
        static OPCODE opLDRBptrplr(), opLDRBofrplr(), opLDRBprrplr();
        static OPCODE opLDRBptrmar(), opLDRBofrmar(), opLDRBprrmar();
        static OPCODE opLDRBptrpar(), opLDRBofrpar(), opLDRBprrpar();
        static OPCODE opLDRBptrmrr(), opLDRBofrmrr(), opLDRBprrmrr();
        static OPCODE opLDRBptrprr(), opLDRBofrprr(), opLDRBprrprr();

        static OPCODE opSTRBptrmll(), opSTRBofrmll(), opSTRBprrmll();
        static OPCODE opSTRBptrpll(), opSTRBofrpll(), opSTRBprrpll();
        static OPCODE opSTRBptrmlr(), opSTRBofrmlr(), opSTRBprrmlr();
        static OPCODE opSTRBptrplr(), opSTRBofrplr(), opSTRBprrplr();
        static OPCODE opSTRBptrmar(), opSTRBofrmar(), opSTRBprrmar();
        static OPCODE opSTRBptrpar(), opSTRBofrpar(), opSTRBprrpar();
        static OPCODE opSTRBptrmrr(), opSTRBofrmrr(), opSTRBprrmrr();
        static OPCODE opSTRBptrprr(), opSTRBofrprr(), opSTRBprrprr();

        static OPCODE opLDRTptrmll(), opLDRTofrmll(), opLDRTprrmll();
        static OPCODE opLDRTptrpll(), opLDRTofrpll(), opLDRTprrpll();
        static OPCODE opLDRTptrmlr(), opLDRTofrmlr(), opLDRTprrmlr();
        static OPCODE opLDRTptrplr(), opLDRTofrplr(), opLDRTprrplr();
        static OPCODE opLDRTptrmar(), opLDRTofrmar(), opLDRTprrmar();
        static OPCODE opLDRTptrpar(), opLDRTofrpar(), opLDRTprrpar();
        static OPCODE opLDRTptrmrr(), opLDRTofrmrr(), opLDRTprrmrr();
        static OPCODE opLDRTptrprr(), opLDRTofrprr(), opLDRTprrprr();

        static OPCODE opSTRTptrmll(), opSTRTofrmll(), opSTRTprrmll();
        static OPCODE opSTRTptrpll(), opSTRTofrpll(), opSTRTprrpll();
        static OPCODE opSTRTptrmlr(), opSTRTofrmlr(), opSTRTprrmlr();
        static OPCODE opSTRTptrplr(), opSTRTofrplr(), opSTRTprrplr();
        static OPCODE opSTRTptrmar(), opSTRTofrmar(), opSTRTprrmar();
        static OPCODE opSTRTptrpar(), opSTRTofrpar(), opSTRTprrpar();
        static OPCODE opSTRTptrmrr(), opSTRTofrmrr(), opSTRTprrmrr();
        static OPCODE opSTRTptrprr(), opSTRTofrprr(), opSTRTprrprr();

        static OPCODE opLDRBTptrmll(), opLDRBTofrmll(), opLDRBTprrmll();
        static OPCODE opLDRBTptrpll(), opLDRBTofrpll(), opLDRBTprrpll();
        static OPCODE opLDRBTptrmlr(), opLDRBTofrmlr(), opLDRBTprrmlr();
        static OPCODE opLDRBTptrplr(), opLDRBTofrplr(), opLDRBTprrplr();
        static OPCODE opLDRBTptrmar(), opLDRBTofrmar(), opLDRBTprrmar();
        static OPCODE opLDRBTptrpar(), opLDRBTofrpar(), opLDRBTprrpar();
        static OPCODE opLDRBTptrmrr(), opLDRBTofrmrr(), opLDRBTprrmrr();
        static OPCODE opLDRBTptrprr(), opLDRBTofrprr(), opLDRBTprrprr();

        static OPCODE opSTRBTptrmll(), opSTRBTofrmll(), opSTRBTprrmll();
        static OPCODE opSTRBTptrpll(), opSTRBTofrpll(), opSTRBTprrpll();
        static OPCODE opSTRBTptrmlr(), opSTRBTofrmlr(), opSTRBTprrmlr();
        static OPCODE opSTRBTptrplr(), opSTRBTofrplr(), opSTRBTprrplr();
        static OPCODE opSTRBTptrmar(), opSTRBTofrmar(), opSTRBTprrmar();
        static OPCODE opSTRBTptrpar(), opSTRBTofrpar(), opSTRBTprrpar();
        static OPCODE opSTRBTptrmrr(), opSTRBTofrmrr(), opSTRBTprrmrr();
        static OPCODE opSTRBTptrprr(), opSTRBTofrprr(), opSTRBTprrprr();

        //---SWP-------------------------------------------------------------------

        static OPCODE opSWP(), opSWPB();

        //---LoaD/STore Multiple; Increment/Decrement After/Before-----------------

        static OPCODE opLDMIA(), opLDMIB(), opLDMDA(), opLDMDB();
        static OPCODE opSTMIA(), opSTMIB(), opSTMDA(), opSTMDB();

        static OPCODE opLDMIAu(), opLDMIBu(), opLDMDAu(), opLDMDBu();
        static OPCODE opSTMIAu(), opSTMIBu(), opSTMDAu(), opSTMDBu();

        static OPCODE opLDMIAw(), opLDMIBw(), opLDMDAw(), opLDMDBw();
        static OPCODE opSTMIAw(), opSTMIBw(), opSTMDAw(), opSTMDBw();

        static OPCODE opLDMIAuw(), opLDMIBuw(), opLDMDAuw(), opLDMDBuw();
        static OPCODE opSTMIAuw(), opSTMIBuw(), opSTMDAuw(), opSTMDBuw();

        //---SoftWare Interrupts---------------------------------------------------

        static OPCODE opSWI();

        //---Thumb ops-------------------------------------------------------------

        static OPCODE topLSLimm(), topLSRimm(), topASRimm();
        static OPCODE topADDreg(), topSUBreg();
        static OPCODE topADDimm3(), topSUBimm3();

        static OPCODE topMOVimm8r0(), topMOVimm8r1();
        static OPCODE topMOVimm8r2(), topMOVimm8r3();
        static OPCODE topMOVimm8r4(), topMOVimm8r5();
        static OPCODE topMOVimm8r6(), topMOVimm8r7();
        static OPCODE topCMPimm8r0(), topCMPimm8r1();
        static OPCODE topCMPimm8r2(), topCMPimm8r3();
        static OPCODE topCMPimm8r4(), topCMPimm8r5();
        static OPCODE topCMPimm8r6(), topCMPimm8r7();
        static OPCODE topADDimm8r0(), topADDimm8r1();
        static OPCODE topADDimm8r2(), topADDimm8r3();
        static OPCODE topADDimm8r4(), topADDimm8r5();
        static OPCODE topADDimm8r6(), topADDimm8r7();
        static OPCODE topSUBimm8r0(), topSUBimm8r1();
        static OPCODE topSUBimm8r2(), topSUBimm8r3();
        static OPCODE topSUBimm8r4(), topSUBimm8r5();
        static OPCODE topSUBimm8r6(), topSUBimm8r7();

        static OPCODE topDPg1(), topDPg2(), topDPg3(), topDPg4();
        static OPCODE topADDH(), topMOVH(), topCMPH(), topBXreg();

        static OPCODE topLDRPCr0(), topLDRPCr1();
        static OPCODE topLDRPCr2(), topLDRPCr3();
        static OPCODE topLDRPCr4(), topLDRPCr5();
        static OPCODE topLDRPCr6(), topLDRPCr7();

        static OPCODE topSTRSPr0(), topSTRSPr1();
        static OPCODE topSTRSPr2(), topSTRSPr3();
        static OPCODE topSTRSPr4(), topSTRSPr5();
        static OPCODE topSTRSPr6(), topSTRSPr7();
        static OPCODE topLDRSPr0(), topLDRSPr1();
        static OPCODE topLDRSPr2(), topLDRSPr3();
        static OPCODE topLDRSPr4(), topLDRSPr5();
        static OPCODE topLDRSPr6(), topLDRSPr7();

        static OPCODE topSTRreg(),  topLDRreg();
        static OPCODE topSTRHreg(), topLDRHreg();
        static OPCODE topSTRBreg(), topLDRBreg();

        static OPCODE topLDRSBreg(), topLDRSHreg();

        static OPCODE topSTRimm5(),  topLDRimm5();
        static OPCODE topSTRBimm5(), topLDRBimm5();
        static OPCODE topSTRHimm5(), topLDRHimm5();

        static OPCODE topADDPCr0(), topADDPCr1();
        static OPCODE topADDPCr2(), topADDPCr3();
        static OPCODE topADDPCr4(), topADDPCr5();
        static OPCODE topADDPCr6(), topADDPCr7();
        static OPCODE topADDSPr0(), topADDSPr1();
        static OPCODE topADDSPr2(), topADDSPr3();
        static OPCODE topADDSPr4(), topADDSPr5();
        static OPCODE topADDSPr6(), topADDSPr7();
        static OPCODE topADDSPimm7();

        static OPCODE topPUSH(), topPUSHlr();
        static OPCODE topPOP(),  topPOPpc();

        static OPCODE topSWI();

        static OPCODE topLDMIAr0(), topLDMIAr1();
        static OPCODE topLDMIAr2(), topLDMIAr3();
        static OPCODE topLDMIAr4(), topLDMIAr5();
        static OPCODE topLDMIAr6(), topLDMIAr7();
        static OPCODE topSTMIAr0(), topSTMIAr1();
        static OPCODE topSTMIAr2(), topSTMIAr3();
        static OPCODE topSTMIAr4(), topSTMIAr5();
        static OPCODE topSTMIAr6(), topSTMIAr7();

        static OPCODE topBEQ(), topBNE(), topBCS(), topBCC();
        static OPCODE topBMI(), topBPL(), topBVS(), topBVC();
        static OPCODE topBHI(), topBLS(), topBGE(), topBLT();
        static OPCODE topBGT(), topBLE();

        static OPCODE topB(), topBLsetup(), topBLoff();

#ifdef ARM7TDMI_X86	
        //---Dynarec instructions

	//---Exceptional opcodes
	static OPCODE xopUNP();   // Unpredictable op
        static OPCODE xopUND();   // Undefined op
        static OPCODE xopUNI();   // Unimplemented op
	static OPCODE xopUNL();   // Unlinked (opcode cache filler)
        static OPCODE xtopUND();  // Thumb-mode undefined
	
        //---Branch (with Link)----------------------------------------------------

        static OPCODE xopB(), xopBL();
        static OPCODE xopBX();

        //---DP (DataProcessing); LsL/LsR/AsR/RoR; Immediate/Register--------------

        static OPCODE xopANDlli(), xopANDllr();
        static OPCODE xopANDlri(), xopANDlrr();
        static OPCODE xopANDari(), xopANDarr();
        static OPCODE xopANDrri(), xopANDrrr();
        static OPCODE xopANDimm();

        static OPCODE xopEORlli(), xopEORllr();
        static OPCODE xopEORlri(), xopEORlrr();
        static OPCODE xopEORari(), xopEORarr();
        static OPCODE xopEORrri(), xopEORrrr();
        static OPCODE xopEORimm();

        static OPCODE xopSUBlli(), xopSUBllr();
        static OPCODE xopSUBlri(), xopSUBlrr();
        static OPCODE xopSUBari(), xopSUBarr();
        static OPCODE xopSUBrri(), xopSUBrrr();
        static OPCODE xopSUBimm();

        static OPCODE xopRSBlli(), xopRSBllr();
        static OPCODE xopRSBlri(), xopRSBlrr();
        static OPCODE xopRSBari(), xopRSBarr();
        static OPCODE xopRSBrri(), xopRSBrrr();
        static OPCODE xopRSBimm();

        static OPCODE xopADDlli(), xopADDllr();
        static OPCODE xopADDlri(), xopADDlrr();
        static OPCODE xopADDari(), xopADDarr();
        static OPCODE xopADDrri(), xopADDrrr();
        static OPCODE xopADDimm();

        static OPCODE xopADClli(), xopADCllr();
        static OPCODE xopADClri(), xopADClrr();
        static OPCODE xopADCari(), xopADCarr();
        static OPCODE xopADCrri(), xopADCrrr();
        static OPCODE xopADCimm();

        static OPCODE xopSBClli(), xopSBCllr();
        static OPCODE xopSBClri(), xopSBClrr();
        static OPCODE xopSBCari(), xopSBCarr();
        static OPCODE xopSBCrri(), xopSBCrrr();
        static OPCODE xopSBCimm();

        static OPCODE xopRSClli(), xopRSCllr();
        static OPCODE xopRSClri(), xopRSClrr();
        static OPCODE xopRSCari(), xopRSCarr();
        static OPCODE xopRSCrri(), xopRSCrrr();
        static OPCODE xopRSCimm();

        static OPCODE xopORRlli(), xopORRllr();
        static OPCODE xopORRlri(), xopORRlrr();
        static OPCODE xopORRari(), xopORRarr();
        static OPCODE xopORRrri(), xopORRrrr();
        static OPCODE xopORRimm();

        static OPCODE xopMOVlli(), xopMOVllr();
        static OPCODE xopMOVlri(), xopMOVlrr();
        static OPCODE xopMOVari(), xopMOVarr();
        static OPCODE xopMOVrri(), xopMOVrrr();
        static OPCODE xopMOVimm();

        static OPCODE xopBIClli(), xopBICllr();
        static OPCODE xopBIClri(), xopBIClrr();
        static OPCODE xopBICari(), xopBICarr();
        static OPCODE xopBICrri(), xopBICrrr();
        static OPCODE xopBICimm();

        static OPCODE xopMVNlli(), xopMVNllr();
        static OPCODE xopMVNlri(), xopMVNlrr();
        static OPCODE xopMVNari(), xopMVNarr();
        static OPCODE xopMVNrri(), xopMVNrrr();
        static OPCODE xopMVNimm();

        static OPCODE xopANDSlli(), xopANDSllr();
        static OPCODE xopANDSlri(), xopANDSlrr();
        static OPCODE xopANDSari(), xopANDSarr();
        static OPCODE xopANDSrri(), xopANDSrrr();
        static OPCODE xopANDSimm();

        static OPCODE xopEORSlli(), xopEORSllr();
        static OPCODE xopEORSlri(), xopEORSlrr();
        static OPCODE xopEORSari(), xopEORSarr();
        static OPCODE xopEORSrri(), xopEORSrrr();
        static OPCODE xopEORSimm();

        static OPCODE xopSUBSlli(), xopSUBSllr();
        static OPCODE xopSUBSlri(), xopSUBSlrr();
        static OPCODE xopSUBSari(), xopSUBSarr();
        static OPCODE xopSUBSrri(), xopSUBSrrr();
        static OPCODE xopSUBSimm();

        static OPCODE xopRSBSlli(), xopRSBSllr();
        static OPCODE xopRSBSlri(), xopRSBSlrr();
        static OPCODE xopRSBSari(), xopRSBSarr();
        static OPCODE xopRSBSrri(), xopRSBSrrr();
        static OPCODE xopRSBSimm();

        static OPCODE xopADDSlli(), xopADDSllr();
        static OPCODE xopADDSlri(), xopADDSlrr();
        static OPCODE xopADDSari(), xopADDSarr();
        static OPCODE xopADDSrri(), xopADDSrrr();
        static OPCODE xopADDSimm();

        static OPCODE xopADCSlli(), xopADCSllr();
        static OPCODE xopADCSlri(), xopADCSlrr();
        static OPCODE xopADCSari(), xopADCSarr();
        static OPCODE xopADCSrri(), xopADCSrrr();
        static OPCODE xopADCSimm();

        static OPCODE xopSBCSlli(), xopSBCSllr();
        static OPCODE xopSBCSlri(), xopSBCSlrr();
        static OPCODE xopSBCSari(), xopSBCSarr();
        static OPCODE xopSBCSrri(), xopSBCSrrr();
        static OPCODE xopSBCSimm();

        static OPCODE xopRSCSlli(), xopRSCSllr();
        static OPCODE xopRSCSlri(), xopRSCSlrr();
        static OPCODE xopRSCSari(), xopRSCSarr();
        static OPCODE xopRSCSrri(), xopRSCSrrr();
        static OPCODE xopRSCSimm();

        static OPCODE xopTSTSlli(), xopTSTSllr();
        static OPCODE xopTSTSlri(), xopTSTSlrr();
        static OPCODE xopTSTSari(), xopTSTSarr();
        static OPCODE xopTSTSrri(), xopTSTSrrr();
        static OPCODE xopTSTSimm();

        static OPCODE xopTEQSlli(), xopTEQSllr();
        static OPCODE xopTEQSlri(), xopTEQSlrr();
        static OPCODE xopTEQSari(), xopTEQSarr();
        static OPCODE xopTEQSrri(), xopTEQSrrr();
        static OPCODE xopTEQSimm();

        static OPCODE xopCMPSlli(), xopCMPSllr();
        static OPCODE xopCMPSlri(), xopCMPSlrr();
        static OPCODE xopCMPSari(), xopCMPSarr();
        static OPCODE xopCMPSrri(), xopCMPSrrr();
        static OPCODE xopCMPSimm();

        static OPCODE xopCMNSlli(), xopCMNSllr();
        static OPCODE xopCMNSlri(), xopCMNSlrr();
        static OPCODE xopCMNSari(), xopCMNSarr();
        static OPCODE xopCMNSrri(), xopCMNSrrr();
        static OPCODE xopCMNSimm();

        static OPCODE xopORRSlli(), xopORRSllr();
        static OPCODE xopORRSlri(), xopORRSlrr();
        static OPCODE xopORRSari(), xopORRSarr();
        static OPCODE xopORRSrri(), xopORRSrrr();
        static OPCODE xopORRSimm();

        static OPCODE xopMOVSlli(), xopMOVSllr();
        static OPCODE xopMOVSlri(), xopMOVSlrr();
        static OPCODE xopMOVSari(), xopMOVSarr();
        static OPCODE xopMOVSrri(), xopMOVSrrr();
        static OPCODE xopMOVSimm();

        static OPCODE xopBICSlli(), xopBICSllr();
        static OPCODE xopBICSlri(), xopBICSlrr();
        static OPCODE xopBICSari(), xopBICSarr();
        static OPCODE xopBICSrri(), xopBICSrrr();
        static OPCODE xopBICSimm();

        static OPCODE xopMVNSlli(), xopMVNSllr();
        static OPCODE xopMVNSlri(), xopMVNSlrr();
        static OPCODE xopMVNSari(), xopMVNSarr();
        static OPCODE xopMVNSrri(), xopMVNSrrr();
        static OPCODE xopMVNSimm();

        //---Multiply/LongMul------------------------------------------------------

        static OPCODE xopMUL(), xopMLA();
        static OPCODE xopUMULL(), xopUMLAL(), xopSMULL(), xopSMLAL();

        static OPCODE xopMULS(), xopMLAS();
        static OPCODE xopUMULLS(), xopUMLALS(), xopSMULLS(), xopSMLALS();

        //---Move Register From Status; Imm/Reg; Cpsr/Spsr-------------------------

        static OPCODE xopMRSrc(), xopMRSrs();

        //---Move Status from Register; Cpsr/Spsr----------------------------------

        static OPCODE xopMSRrc(), xopMSRrs();
        static OPCODE xopMSRic(), xopMSRis();

        //---LS (Load/Store); PReidx/OFfset/PsTidx; Imm/Reg; Minus/Plus; (Shift)---

        static OPCODE xopLDRptim(), xopLDRofim(), xopLDRprim();
        static OPCODE xopLDRptip(), xopLDRofip(), xopLDRprip();

        static OPCODE xopSTRptim(), xopSTRofim(), xopSTRprim();
        static OPCODE xopSTRptip(), xopSTRofip(), xopSTRprip();

        static OPCODE xopLDRBptim(), xopLDRBofim(), xopLDRBprim();
        static OPCODE xopLDRBptip(), xopLDRBofip(), xopLDRBprip();

        static OPCODE xopSTRBptim(), xopSTRBofim(), xopSTRBprim();
        static OPCODE xopSTRBptip(), xopSTRBofip(), xopSTRBprip();

        static OPCODE xopLDRHptim(), xopLDRHofim(), xopLDRHprim();
        static OPCODE xopLDRHptip(), xopLDRHofip(), xopLDRHprip();
        static OPCODE xopLDRHptrm(), xopLDRHofrm(), xopLDRHprrm();
        static OPCODE xopLDRHptrp(), xopLDRHofrp(), xopLDRHprrp();

        static OPCODE xopSTRHptim(), xopSTRHofim(), xopSTRHprim();
        static OPCODE xopSTRHptip(), xopSTRHofip(), xopSTRHprip();
        static OPCODE xopSTRHptrm(), xopSTRHofrm(), xopSTRHprrm();
        static OPCODE xopSTRHptrp(), xopSTRHofrp(), xopSTRHprrp();

        static OPCODE xopLDRTptim(), xopLDRTofim(), xopLDRTprim();
        static OPCODE xopLDRTptip(), xopLDRTofip(), xopLDRTprip();

        static OPCODE xopSTRTptim(), xopSTRTofim(), xopSTRTprim();
        static OPCODE xopSTRTptip(), xopSTRTofip(), xopSTRTprip();

        static OPCODE xopLDRBTptim(), xopLDRBTofim(), xopLDRBTprim();
        static OPCODE xopLDRBTptip(), xopLDRBTofip(), xopLDRBTprip();

        static OPCODE xopSTRBTptim(), xopSTRBTofim(), xopSTRBTprim();
        static OPCODE xopSTRBTptip(), xopSTRBTofip(), xopSTRBTprip();

        static OPCODE xopLDRSHptim(), xopLDRSHofim(), xopLDRSHprim();
        static OPCODE xopLDRSHptip(), xopLDRSHofip(), xopLDRSHprip();
        static OPCODE xopLDRSHptrm(), xopLDRSHofrm(), xopLDRSHprrm();
        static OPCODE xopLDRSHptrp(), xopLDRSHofrp(), xopLDRSHprrp();

        static OPCODE xopLDRSBptim(), xopLDRSBofim(), xopLDRSBprim();
        static OPCODE xopLDRSBptip(), xopLDRSBofip(), xopLDRSBprip();
        static OPCODE xopLDRSBptrm(), xopLDRSBofrm(), xopLDRSBprrm();
        static OPCODE xopLDRSBptrp(), xopLDRSBofrp(), xopLDRSBprrp();

        static OPCODE xopLDRptrmll(), xopLDRofrmll(), xopLDRprrmll();
        static OPCODE xopLDRptrpll(), xopLDRofrpll(), xopLDRprrpll();
        static OPCODE xopLDRptrmlr(), xopLDRofrmlr(), xopLDRprrmlr();
        static OPCODE xopLDRptrplr(), xopLDRofrplr(), xopLDRprrplr();
        static OPCODE xopLDRptrmar(), xopLDRofrmar(), xopLDRprrmar();
        static OPCODE xopLDRptrpar(), xopLDRofrpar(), xopLDRprrpar();
        static OPCODE xopLDRptrmrr(), xopLDRofrmrr(), xopLDRprrmrr();
        static OPCODE xopLDRptrprr(), xopLDRofrprr(), xopLDRprrprr();

        static OPCODE xopSTRptrmll(), xopSTRofrmll(), xopSTRprrmll();
        static OPCODE xopSTRptrpll(), xopSTRofrpll(), xopSTRprrpll();
        static OPCODE xopSTRptrmlr(), xopSTRofrmlr(), xopSTRprrmlr();
        static OPCODE xopSTRptrplr(), xopSTRofrplr(), xopSTRprrplr();
        static OPCODE xopSTRptrmar(), xopSTRofrmar(), xopSTRprrmar();
        static OPCODE xopSTRptrpar(), xopSTRofrpar(), xopSTRprrpar();
        static OPCODE xopSTRptrmrr(), xopSTRofrmrr(), xopSTRprrmrr();
        static OPCODE xopSTRptrprr(), xopSTRofrprr(), xopSTRprrprr();

        static OPCODE xopLDRBptrmll(), xopLDRBofrmll(), xopLDRBprrmll();
        static OPCODE xopLDRBptrpll(), xopLDRBofrpll(), xopLDRBprrpll();
        static OPCODE xopLDRBptrmlr(), xopLDRBofrmlr(), xopLDRBprrmlr();
        static OPCODE xopLDRBptrplr(), xopLDRBofrplr(), xopLDRBprrplr();
        static OPCODE xopLDRBptrmar(), xopLDRBofrmar(), xopLDRBprrmar();
        static OPCODE xopLDRBptrpar(), xopLDRBofrpar(), xopLDRBprrpar();
        static OPCODE xopLDRBptrmrr(), xopLDRBofrmrr(), xopLDRBprrmrr();
        static OPCODE xopLDRBptrprr(), xopLDRBofrprr(), xopLDRBprrprr();

        static OPCODE xopSTRBptrmll(), xopSTRBofrmll(), xopSTRBprrmll();
        static OPCODE xopSTRBptrpll(), xopSTRBofrpll(), xopSTRBprrpll();
        static OPCODE xopSTRBptrmlr(), xopSTRBofrmlr(), xopSTRBprrmlr();
        static OPCODE xopSTRBptrplr(), xopSTRBofrplr(), xopSTRBprrplr();
        static OPCODE xopSTRBptrmar(), xopSTRBofrmar(), xopSTRBprrmar();
        static OPCODE xopSTRBptrpar(), xopSTRBofrpar(), xopSTRBprrpar();
        static OPCODE xopSTRBptrmrr(), xopSTRBofrmrr(), xopSTRBprrmrr();
        static OPCODE xopSTRBptrprr(), xopSTRBofrprr(), xopSTRBprrprr();

        static OPCODE xopLDRTptrmll(), xopLDRTofrmll(), xopLDRTprrmll();
        static OPCODE xopLDRTptrpll(), xopLDRTofrpll(), xopLDRTprrpll();
        static OPCODE xopLDRTptrmlr(), xopLDRTofrmlr(), xopLDRTprrmlr();
        static OPCODE xopLDRTptrplr(), xopLDRTofrplr(), xopLDRTprrplr();
        static OPCODE xopLDRTptrmar(), xopLDRTofrmar(), xopLDRTprrmar();
        static OPCODE xopLDRTptrpar(), xopLDRTofrpar(), xopLDRTprrpar();
        static OPCODE xopLDRTptrmrr(), xopLDRTofrmrr(), xopLDRTprrmrr();
        static OPCODE xopLDRTptrprr(), xopLDRTofrprr(), xopLDRTprrprr();

        static OPCODE xopSTRTptrmll(), xopSTRTofrmll(), xopSTRTprrmll();
        static OPCODE xopSTRTptrpll(), xopSTRTofrpll(), xopSTRTprrpll();
        static OPCODE xopSTRTptrmlr(), xopSTRTofrmlr(), xopSTRTprrmlr();
        static OPCODE xopSTRTptrplr(), xopSTRTofrplr(), xopSTRTprrplr();
        static OPCODE xopSTRTptrmar(), xopSTRTofrmar(), xopSTRTprrmar();
        static OPCODE xopSTRTptrpar(), xopSTRTofrpar(), xopSTRTprrpar();
        static OPCODE xopSTRTptrmrr(), xopSTRTofrmrr(), xopSTRTprrmrr();
        static OPCODE xopSTRTptrprr(), xopSTRTofrprr(), xopSTRTprrprr();

        static OPCODE xopLDRBTptrmll(), xopLDRBTofrmll(), xopLDRBTprrmll();
        static OPCODE xopLDRBTptrpll(), xopLDRBTofrpll(), xopLDRBTprrpll();
        static OPCODE xopLDRBTptrmlr(), xopLDRBTofrmlr(), xopLDRBTprrmlr();
        static OPCODE xopLDRBTptrplr(), xopLDRBTofrplr(), xopLDRBTprrplr();
        static OPCODE xopLDRBTptrmar(), xopLDRBTofrmar(), xopLDRBTprrmar();
        static OPCODE xopLDRBTptrpar(), xopLDRBTofrpar(), xopLDRBTprrpar();
        static OPCODE xopLDRBTptrmrr(), xopLDRBTofrmrr(), xopLDRBTprrmrr();
        static OPCODE xopLDRBTptrprr(), xopLDRBTofrprr(), xopLDRBTprrprr();

        static OPCODE xopSTRBTptrmll(), xopSTRBTofrmll(), xopSTRBTprrmll();
        static OPCODE xopSTRBTptrpll(), xopSTRBTofrpll(), xopSTRBTprrpll();
        static OPCODE xopSTRBTptrmlr(), xopSTRBTofrmlr(), xopSTRBTprrmlr();
        static OPCODE xopSTRBTptrplr(), xopSTRBTofrplr(), xopSTRBTprrplr();
        static OPCODE xopSTRBTptrmar(), xopSTRBTofrmar(), xopSTRBTprrmar();
        static OPCODE xopSTRBTptrpar(), xopSTRBTofrpar(), xopSTRBTprrpar();
        static OPCODE xopSTRBTptrmrr(), xopSTRBTofrmrr(), xopSTRBTprrmrr();
        static OPCODE xopSTRBTptrprr(), xopSTRBTofrprr(), xopSTRBTprrprr();

        //---SWP-------------------------------------------------------------------

        static OPCODE xopSWP(), xopSWPB();

        //---LoaD/STore Multiple; Increment/Decrement After/Before-----------------

        static OPCODE xopLDMIA(), xopLDMIB(), xopLDMDA(), xopLDMDB();
        static OPCODE xopSTMIA(), xopSTMIB(), xopSTMDA(), xopSTMDB();

        static OPCODE xopLDMIAu(), xopLDMIBu(), xopLDMDAu(), xopLDMDBu();
        static OPCODE xopSTMIAu(), xopSTMIBu(), xopSTMDAu(), xopSTMDBu();

        static OPCODE xopLDMIAw(), xopLDMIBw(), xopLDMDAw(), xopLDMDBw();
        static OPCODE xopSTMIAw(), xopSTMIBw(), xopSTMDAw(), xopSTMDBw();

        static OPCODE xopLDMIAuw(), xopLDMIBuw(), xopLDMDAuw(), xopLDMDBuw();
        static OPCODE xopSTMIAuw(), xopSTMIBuw(), xopSTMDAuw(), xopSTMDBuw();

        //---SoftWare Interrupts---------------------------------------------------

        static OPCODE xopSWI();

        //---Thumb xops-------------------------------------------------------------

        static OPCODE xtopLSLimm(), xtopLSRimm(), xtopASRimm();
        static OPCODE xtopADDreg(), xtopSUBreg();
        static OPCODE xtopADDimm3(), xtopSUBimm3();

        static OPCODE xtopMOVimm8r0(), xtopMOVimm8r1();
        static OPCODE xtopMOVimm8r2(), xtopMOVimm8r3();
        static OPCODE xtopMOVimm8r4(), xtopMOVimm8r5();
        static OPCODE xtopMOVimm8r6(), xtopMOVimm8r7();
        static OPCODE xtopCMPimm8r0(), xtopCMPimm8r1();
        static OPCODE xtopCMPimm8r2(), xtopCMPimm8r3();
        static OPCODE xtopCMPimm8r4(), xtopCMPimm8r5();
        static OPCODE xtopCMPimm8r6(), xtopCMPimm8r7();
        static OPCODE xtopADDimm8r0(), xtopADDimm8r1();
        static OPCODE xtopADDimm8r2(), xtopADDimm8r3();
        static OPCODE xtopADDimm8r4(), xtopADDimm8r5();
        static OPCODE xtopADDimm8r6(), xtopADDimm8r7();
        static OPCODE xtopSUBimm8r0(), xtopSUBimm8r1();
        static OPCODE xtopSUBimm8r2(), xtopSUBimm8r3();
        static OPCODE xtopSUBimm8r4(), xtopSUBimm8r5();
        static OPCODE xtopSUBimm8r6(), xtopSUBimm8r7();

        static OPCODE xtopDPg1(), xtopDPg2(), xtopDPg3(), xtopDPg4();
        static OPCODE xtopADDH(), xtopMOVH(), xtopCMPH(), xtopBXreg();

        static OPCODE xtopLDRPCr0(), xtopLDRPCr1();
        static OPCODE xtopLDRPCr2(), xtopLDRPCr3();
        static OPCODE xtopLDRPCr4(), xtopLDRPCr5();
        static OPCODE xtopLDRPCr6(), xtopLDRPCr7();

        static OPCODE xtopSTRSPr0(), xtopSTRSPr1();
        static OPCODE xtopSTRSPr2(), xtopSTRSPr3();
        static OPCODE xtopSTRSPr4(), xtopSTRSPr5();
        static OPCODE xtopSTRSPr6(), xtopSTRSPr7();
        static OPCODE xtopLDRSPr0(), xtopLDRSPr1();
        static OPCODE xtopLDRSPr2(), xtopLDRSPr3();
        static OPCODE xtopLDRSPr4(), xtopLDRSPr5();
        static OPCODE xtopLDRSPr6(), xtopLDRSPr7();

        static OPCODE xtopSTRreg(),  xtopLDRreg();
        static OPCODE xtopSTRHreg(), xtopLDRHreg();
        static OPCODE xtopSTRBreg(), xtopLDRBreg();

        static OPCODE xtopLDRSBreg(), xtopLDRSHreg();

        static OPCODE xtopSTRimm5(),  xtopLDRimm5();
        static OPCODE xtopSTRBimm5(), xtopLDRBimm5();
        static OPCODE xtopSTRHimm5(), xtopLDRHimm5();

        static OPCODE xtopADDPCr0(), xtopADDPCr1();
        static OPCODE xtopADDPCr2(), xtopADDPCr3();
        static OPCODE xtopADDPCr4(), xtopADDPCr5();
        static OPCODE xtopADDPCr6(), xtopADDPCr7();
        static OPCODE xtopADDSPr0(), xtopADDSPr1();
        static OPCODE xtopADDSPr2(), xtopADDSPr3();
        static OPCODE xtopADDSPr4(), xtopADDSPr5();
        static OPCODE xtopADDSPr6(), xtopADDSPr7();
        static OPCODE xtopADDSPimm7();

        static OPCODE xtopPUSH(), xtopPUSHlr();
        static OPCODE xtopPOP(),  xtopPOPpc();

        static OPCODE xtopSWI();

        static OPCODE xtopLDMIAr0(), xtopLDMIAr1();
        static OPCODE xtopLDMIAr2(), xtopLDMIAr3();
        static OPCODE xtopLDMIAr4(), xtopLDMIAr5();
        static OPCODE xtopLDMIAr6(), xtopLDMIAr7();
        static OPCODE xtopSTMIAr0(), xtopSTMIAr1();
        static OPCODE xtopSTMIAr2(), xtopSTMIAr3();
        static OPCODE xtopSTMIAr4(), xtopSTMIAr5();
        static OPCODE xtopSTMIAr6(), xtopSTMIAr7();

        static OPCODE xtopBEQ(), xtopBNE(), xtopBCS(), xtopBCC();
        static OPCODE xtopBMI(), xtopBPL(), xtopBVS(), xtopBVC();
        static OPCODE xtopBHI(), xtopBLS(), xtopBGE(), xtopBLT();
        static OPCODE xtopBGT(), xtopBLE();

        static OPCODE xtopB(), xtopBLsetup(), xtopBLoff();
#endif

};

// Allow opcode implementation files common to both ARMs
#define OPC ARM7TDMI::OPCODE ARM7TDMI::

#define ERR_CPU_DYNALLOC 0x0111

#endif//__ARM7TDMI_H_

/*** EOF: arm7tdmi.h *****************************************************/

